/*
    webalizer - a web server log analysis program

    Copyright (C) 1997-2001  Bradford L. Barrett (brad@mrunix.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version, and provided that the above
    copyright and permission notice is included with all distributed
    copies of this or derived software.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA

    This software uses the gd graphics library, which is copyright by
    Quest Protein Database Center, Cold Spring Harbor Labs.  Please
    see the documentation supplied with the library for additional
    information and license terms, or visit www.boutell.com/gd/ for the
    most recent version of the library and supporting documentation.
*/

/*********************************************/
/* STANDARD INCLUDES                         */
/*********************************************/

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>                           /* normal stuff             */
#include <ctype.h>
#include <sys/utsname.h>
#include <sys/times.h>
#include <libgen.h>
#include <pcre/pcre.h>

/* ensure getopt */
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif

/* ensure sys/types */
#ifndef _SYS_TYPES_H
#include <sys/types.h>
#endif

/* some systems need this */
#ifdef HAVE_MATH_H
#include <math.h>
#endif

/* SunOS 4.x Fix */
#ifndef CLK_TCK
#define CLK_TCK _SC_CLK_TCK
#endif

#include "webalizer.h"                        /* main header              */
#include "lang.h"
#include "parser.h"

/* internal function prototypes */
void fmt_logrec(char *);
static int  parse_record_web(char *);
static int  parse_record_ftp(char *);
static int  parse_record_squid(char *);
static void form_url_dirname(char *, char *);

/*
 * pre-compiled Perl-like regular expressions
 */
struct pcre_set {
    char *pattern;	/* regular expression, or NULL ==> end of list */
    pcre *re;		/* compiled pattern */
    pcre_extra *pe;	/* optimized compiled pattern */
};
/* keep the pc_re enum in the same order as the pc[] array */
enum pc_re {
    RE_GET_URL=0,	/* get URL our of an HTTP command */
    RE_STRIP_ARGS,	/* strip off HTTP args and HTTP command fields */
    RE_UN_SLASHLASH,	/* convert //'s into / */
    RE_SLASHDOT,	/* convert /./'s in / */
    RE_UN_DOTDOT,	/* remove foo/../ where foo doesn't end in . */
    RE_TRAIL_DOTDOT,	/* convert trialing /foo/.. into / */
    RE_LEAD_DOTDOT,	/* remove leading /../'s and /./'s */
    RE_UN_DOTDOT2,	/* remove foo./../ */
};
/* keep the pc_re enum in the same order as the pc[] array */
static struct pcre_set pc[] = {

    /* RE_GET_URL */
    {"^\"?\\S+\\s+(\\S+)(\\s.*)?\"?$", NULL, NULL},

    /* RE_STRIP_ARGS */
    {"^([^?\\s]+)[?\\s]", NULL, NULL},

    /* RE_UN_SLASHLASH */
    {"^(.*?)//+(.*)$", NULL, NULL},

    /* RE_SLASHDOT */
    {"^(.*?/)(\\./)+(.*)$", NULL, NULL},

    /* RE_UN_DOTDOT */
    {"^(.*/)?[^/.]+/\\.\\./(.*)$", NULL, NULL},

    /* RE_TRAIL_DOTDOT */
    {"^(.*/)\\.\\.$", NULL, NULL},

    /* RE_LEAD_DOTDOT */
    {"^(/\\.{1,2})+(/.*)$", NULL, NULL},

    /* RE_UN_DOTDOT2 */
    {"^(.*/)?[^/]+/\\.\\./(.*)$", NULL, NULL},

    {NULL, NULL, NULL}	/* must be last */
};
static int pcre_ready = 0;	/* 1 ==> regular expressions pre-compiled */
#define OV_LEN (3*16)   /* substring info vector len - must be multiple of 3 */
static int pcredbg = 0;	/* 1 ==> output debugging to stderr */


/******************************************************/
/* PCRE_SETUP - compile Perl-like regular expressions */
/******************************************************/

static void pcre_setup(void)
{
    const char *comp_err;	/* regular expression compile error */
    int comp_offset;		/* offset in pattern where error occured */
    int i;

    /* firewall - no need to setup twice */
    if (pcre_ready != 0) {
	return;
    }

    /* compile each regular expression */
    for (i=0; pc[i].pattern != NULL; ++i) {

	/* compile the regular expression pattern */
	pc[i].re = pcre_compile(pc[i].pattern, 0, &comp_err,
				&comp_offset, NULL);
	if (pc[i].re == NULL) {
	    fprintf(stderr,
	        "FATAL: regular expression %d error at offset %d: %s\n",
	        i, comp_offset, comp_err);
	    fprintf(stderr, "FATAL: bad reexp: %s\n", pc[i].pattern);
	    exit(1);
	}

	/* optimize compiled regular expression */
	pc[i].pe = pcre_study(pc[i].re, 0, &comp_err);
	if (comp_err != NULL) {
	    fprintf(stderr,
	    	    "FATAL: regular expression %d optimize failed: %s\n",
		    i, comp_err);
	    exit(2);
	}
    }

    /* everything setup */
    pcre_ready = 1;
    return;
}


/*********************************************/
/* FMT_LOGREC - terminate log fields w/zeros */
/*********************************************/

void fmt_logrec(char *buffer)
{
   char *cp=buffer;
   int  q=0,b=0,p=0;

   while (*cp != '\0')
   {
      /* break record up, terminate fields with '\0' */
      switch (*cp)
      {
       case ' ': if (b || q || p) break; *cp='\0'; break;
       case '"': q^=1;  break;
       case '[': if (q) break; b++; break;
       case ']': if (q) break; if (b>0) b--; break;
       case '(': if (q) break; p++; break;
       case ')': if (q) break; if (p>0) p--; break;
      }
      cp++;
   }
}

/*********************************************/
/* PARSE_RECORD - uhhh, you know...          */
/*********************************************/

int parse_record(char *buffer)
{
   int ret;	/* return status */

   /* setup regular expressions if needed */
   if ((ntop_dirs || ntop_dirsK || dump_dirs) && pcre_ready == 0) {
      pcre_setup();
   }

   /* clear out structure */
   memset(&log_rec,0,sizeof(struct log_struct));
/*
   log_rec.hostname[0]=0;
   log_rec.datetime[0]=0;
   log_rec.url[0]=0;
   log_rec.dir[0]=0;
   log_rec.resp_code=0;
   log_rec.xfer_size=0;
   log_rec.refer[0]=0;
   log_rec.agent[0]=0;
   log_rec.srchstr[0]=0;
   log_rec.ident[0]=0;
*/
#ifdef USE_DNS
   memset(&log_rec.addr,0,sizeof(struct in_addr));
#endif

   /* call appropriate handler */
   switch (log_type)
   {
      default:
      case LOG_CLF:   ret = parse_record_web(buffer);   break; /* clf   */
      case LOG_FTP:   ret = parse_record_ftp(buffer);   break; /* ftp   */
      case LOG_SQUID: ret = parse_record_squid(buffer); break; /* squid */
   }

   /* form the dir name of the URL path component */
   if (ntop_dirs || ntop_dirsK || dump_dirs) {
       form_url_dirname(log_rec.url, log_rec.dir);
   }

   /* return status */
   return ret;
}

/*********************************************/
/* PARSE_RECORD_FTP - ftp log handler        */
/*********************************************/

static int parse_record_ftp(char *buffer)
{
   int size;
   int i,j;
   char *cp1, *cp2, *cpx, *cpy, *eob;

   size = strlen(buffer);                 /* get length of buffer        */
   eob = buffer+size;                     /* calculate end of buffer     */
   fmt_logrec(buffer);                    /* seperate fields with \0's   */

   /* Start out with date/time       */
   cp1=buffer;
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   cpx=cp1;       /* save month name */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   i=atoi(cp1);   /* get day number  */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   cpy=cp1;       /* get timestamp   */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   j=atoi(cp1);   /* get year        */

   /* minimal sanity check */
   if (*(cpy+2)!=':' || *(cpy+5)!=':') return -1;
   if (j<1990 || j>2100) return -1;
   if (i<1 || i>31) return -1;

   /* format date/time field         */
   sprintf(log_rec.datetime,"[%02d/%s/%4d:%s -0000]",i,cpx,j,cpy);

   /* skip seconds... */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;

   /* get hostname */
   cp2=log_rec.hostname;
   strncpy(cp2,cp1,MAXHOST-1);

   /* get filesize */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   if (*cp1<'0'||*cp1>'9') log_rec.xfer_size=0;
   else log_rec.xfer_size = strtoul(cp1,NULL,10);

   /* URL stuff */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   cpx=cp1;
   /* get next field for later */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   if (strlen(cpx)>MAXURL-20) *(cpx+(MAXURL-20))=0;

   /* skip next two */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;

   /* fabricate an appropriate request string based on direction */
   if (*cp1=='i') sprintf(log_rec.url,"\"POST %s HTTP/1.0\"",cpx);
      else        sprintf(log_rec.url,"\"GET %s HTTP/1.0\"",cpx);

   if (cp1<eob) cp1++;
   if (cp1<eob) cp1++;
   while (*cp1!=0 && cp1<eob) cp1++;
   if (cp1<eob) cp1++;
   cp2=log_rec.ident;
   while (*cp1!=0 && cp1<eob) *cp2++ = *cp1++;
   *cp2='\0';

   /* return appropriate response code */
   log_rec.resp_code=(*(eob-2)=='i')?206:200;

   return 0;
}

/*********************************************/
/* PARSE_RECORD_WEB - web log handler        */
/*********************************************/

static int parse_record_web(char *buffer)
{
   int size;
   char *cp1, *cp2, *cpx, *eob, *eos;
#if defined(MODIFY_LOG_LINES)
   char *p;
#endif /* MODIFY_LOG_LINES */

   size = strlen(buffer);                 /* get length of buffer        */
   eob = buffer+size;                     /* calculate end of buffer     */
   fmt_logrec(buffer);                    /* seperate fields with \0's   */

   /* HOSTNAME */
   cp1 = cpx = buffer; cp2=log_rec.hostname;
   eos = (cp1+MAXHOST)-1;
   if (eos >= eob) eos=eob-1;

   while ( (*cp1 != '\0') && (cp1 != eos) ) *cp2++ = *cp1++;
   *cp2 = '\0';
   if (*cp1 != '\0')
   {
      if (debug_mode)
      {
         fprintf(stderr,"%s",msg_big_host);
         fprintf(stderr,": %s\n",cpx);
      }
      while (*cp1 != '\0') cp1++;
   }
   if (cp1 < eob) cp1++;

   /* skip next field (ident) */
   while ( (*cp1 != '\0') && (cp1 < eob) ) cp1++;
   if (cp1 < eob) cp1++;

   /* IDENT (authuser) field */
   cpx = cp1;
   cp2 = log_rec.ident;
   eos = (cp1+MAXIDENT-1);
   if (eos >= eob) eos=eob-1;

   while ( (*cp1 != '[') && (cp1 < eos) ) /* remove embeded spaces */
   {
      if (*cp1=='\0') *cp1=' ';
      *cp2++=*cp1++;
   }
   *cp2--='\0';

   if (cp1 >= eob) return -1;

   /* check if oversized username */
   if (*cp1 != '[')
   {
      if (debug_mode)
      {
         fprintf(stderr,"%s",msg_big_user);
         fprintf(stderr,": %s\n",cpx);
      }
      while ( (*cp1 != '[') && (cp1 < eob) ) cp1++;
   }

   /* strip trailing space(s) */
   while (*cp2==' ') *cp2--='\0';

   /* date/time string */
   cpx = cp1;
   cp2 = log_rec.datetime;
   eos = (cp1+28);
   if (eos >= eob) eos=eob-1;

   while ( (*cp1 != '\0') && (cp1 != eos) ) *cp2++ = *cp1++;
   *cp2 = '\0';
   if (*cp1 != '\0')
   {
      if (debug_mode)
      {
         fprintf(stderr,"%s",msg_big_date);
         fprintf(stderr,": %s\n",cpx);
      }
      while (*cp1 != '\0') cp1++;
   }
   if (cp1 < eob) cp1++;

   /* minimal sanity check on timestamp */
   if ( (log_rec.datetime[0] != '[') ||
        (log_rec.datetime[3] != '/') ||
        (cp1 >= eob))  return -1;

   /* HTTP request */
   cpx = cp1;
   cp2 = log_rec.url;
   eos = (cp1+MAXURL-1);
   if (eos >= eob) eos = eob-1;

   while ( (*cp1 != '\0') && (cp1 != eos) ) *cp2++ = *cp1++;
   *cp2 = '\0';
   if (*cp1 != '\0')
   {
      if (debug_mode)
      {
         fprintf(stderr,"%s",msg_big_req);
         fprintf(stderr,": %s\n",cpx);
      }
      while (*cp1 != '\0') cp1++;
   }
   if (cp1 < eob) cp1++;

   if ( (log_rec.url[0] != '"') ||
        (cp1 >= eob) ) return -1;

   if ( (log_rec.url[1] == '"') ||
        (log_rec.url[1] == '-' && log_rec.url[2] == '"') ) {
	return 1;	/* silent ignore of an empty record */
   }

#if defined(MODIFY_LOG_LINES)
   /* -Z means to ignore any record that does not have an absolute URL */
   if (ignore_rel_url) {

	/* look past the "HTTP command */
	for (p = log_rec.url; *p && *p != ' '; ++p) {
	}
	if (*p == ' ') {

	    /* look past the initial "HTTP command for the 1st non-alphanum */
	    for (++p; *p && isascii(*p) && isalnum(*p); ++p) {
	    }

	    /* if we are at a :/ then we have an absolute URL */
	    if (*p != ':' || *(p+1) != '/') {

		/* URL is not of the form alphanum:/... */
		return 1;	/* silent ignore of non abs record */
	    }

	} else {
	    /* no leading "HTTP command */
	    return -1;
	}
   }

   /* -z means to strip off any method://host.name from the URL */
   if (strip_abs_url) {
	char *beyond = log_rec.url;
	char *start;

	/* look past the "HTTP command */
	for (start = beyond; *start && *start != ' '; ++start) {
	}
	if (*start == ' ') {

	    /* look past the HTTP command for the first non-alphanumeric */
	    for (p = ++start; *p && isascii(*p) && isalnum(*p); ++p) {
	    }

	    /* if that character is a :, then we have absulute URL to strip */
	    if (*p == ':') {

		/* we will beyond the method: at a minimum */
		beyond = p+1;

		/* must now strip off // after the method: */
		if (beyond[0] == '/' && beyond[1] == '/') {

		    /* beyond through the method:// */
		    beyond += 2;

		    /* skip the host until the next / */
		    for (p = beyond; *p && *p != '/'; ++p) {
		    }
		    if (*p == '/') {
			/* beyond through the method://host.name */
			beyond = p;
		    }
		}

		/* shift the URL down and this remove the method://host.name */
		p = start;
		while (*beyond) {
		    *p++ = *beyond++;
		}
		*p = '\0';
	    }

	/* must hav a space after the "HTTP command */
	} else {
	    /* no leading "HTTP command */
	    return -1;
	}
   }
#endif /* MODIFY_LOG_LINES */

   /* response code */
   log_rec.resp_code = atoi(cp1);

   /* xfer size */
   while ( (*cp1 != '\0') && (cp1 < eob) ) cp1++;
   if (cp1 < eob) cp1++;
   if (*cp1<'0'||*cp1>'9') log_rec.xfer_size=0;
   else log_rec.xfer_size = strtoul(cp1,NULL,10);

   /* done with CLF record */
   if (cp1>=eob) return 0;

   while ( (*cp1 != '\0') && (*cp1 != '\n') && (cp1 < eob) ) cp1++;
   if (cp1 < eob) cp1++;
   /* get referrer if present */
   cpx = cp1;
   cp2 = log_rec.refer;
   eos = (cp1+MAXREF-1);
   if (eos >= eob) eos = eob-1;

   while ( (*cp1 != '\0') && (*cp1 != '\n') && (cp1 != eos) ) *cp2++ = *cp1++;
   *cp2 = '\0';
   if (*cp1 != '\0')
   {
      if (debug_mode)
      {
         fprintf(stderr,"%s",msg_big_ref);
         fprintf(stderr,": %s\n",cpx);
      }
      while (*cp1 != '\0') cp1++;
   }
   if (cp1 < eob) cp1++;

   cpx = cp1;
   cp2 = log_rec.agent;
   eos = cp1+(MAXAGENT-1);
   if (eos >= eob) eos = eob-1;

   while ( (*cp1 != '\0') && (cp1 != eos) ) *cp2++ = *cp1++;
   *cp2 = '\0';

   return 0;     /* maybe a valid record, return with TRUE */
}

/*********************************************/
/* PARSE_RECORD_SQUID - squid log handler    */
/*********************************************/

static int parse_record_squid(char *buffer)
{
   int size;
   time_t i;
   char *cp1, *cp2, *cpx, *eob, *eos;

   size = strlen(buffer);                 /* get length of buffer        */
   eob = buffer+size;                     /* calculate end of buffer     */
   fmt_logrec(buffer);                    /* seperate fields with \0's   */

   /* date/time */
   cp1=buffer;
   i=atoi(cp1);		/* get timestamp */

   /* format date/time field */
   strftime(log_rec.datetime,sizeof(log_rec.datetime),
            "[%d/%b/%Y:%H:%M:%S -0000]",localtime(&i));

   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;

   /* skip request size */
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;

   /* HOSTNAME */
   cpx = cp1; cp2=log_rec.hostname;
   eos = (cp1+MAXHOST)-1;
   if (eos >= eob) eos=eob-1;

   while ((*cp1 != '\0') && (cp1 != eos)) *cp2++ = *cp1++;
   *cp2='\0';
   if (*cp1 != '\0')
   {
      if (debug_mode)
      {
         fprintf(stderr,"%s",msg_big_host);
         fprintf(stderr,": %s\n",cpx);
      }
      while (*cp1 != '\0') cp1++;
   }
   if (cp1 < eob) cp1++;

   /* skip cache status */
   while (*cp1!=0 && cp1<eob && *cp1!='/') cp1++;
   cp1++;

   /* response code */
   log_rec.resp_code = atoi(cp1);
   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;

   /* xfer size */
   if (*cp1<'0'||*cp1>'9') log_rec.xfer_size=0;
   else log_rec.xfer_size = strtoul(cp1,NULL,10);

   while (*cp1!=0 && cp1<eob) cp1++;
   while (*cp1==0) cp1++;

   /* HTTP request type */
   cpx = cp1;
   cp2 = log_rec.url;
   *cp2++ = '\"';
   eos = (cp1+MAXURL-1);
   if (eos >= eob) eos = eob-1;

   while ( (*cp1 != '\0') && (cp1 != eos) ) *cp2++ = *cp1++;
   *cp2 = '\0';
   if (*cp1 != '\0')
   {
      if (debug_mode)
      {
         fprintf(stderr,"%s",msg_big_req);
         fprintf(stderr,": %s\n",cpx);
      }
      while (*cp1 != '\0') cp1++;
   }
   if (cp1 < eob) cp1++;

   *cp2++ = ' ';

   /* HTTP URL requested */
   cpx = cp1;
   eos = (cp1+MAXURL-1);
   if (eos >= eob) eos = eob-1;

   while ( (*cp1 != '\0') && (cp1 != eos) ) *cp2++ = *cp1++;
   *cp2 = '\0';
   if (*cp1 != '\0')
   {
      if (debug_mode)
      {
         fprintf(stderr,"%s",msg_big_req);
         fprintf(stderr,": %s\n",cpx);
      }
      while (*cp1 != '\0') cp1++;
   }
   if (cp1 < eob) cp1++;

   *cp2++ = '\"';

   /* IDENT (authuser) field */
   cpx = cp1;
   cp2 = log_rec.ident;
   eos = (cp1+MAXIDENT-1);
   if (eos >= eob) eos=eob-1;

   while (*cp1 == ' ') cp1++; /* skip white space */

   while ( (*cp1 != ' ' && *cp1!='\0') && (cp1 < eos) )  *cp2++=*cp1++;

   *cp2--='\0';

   if (cp1 >= eob) return -1;

   /* strip trailing space(s) */
   while (*cp2==' ') *cp2--='\0';

   /* we have no interest in the remaining fields */
   return 0;
}

/*****************************************************************/
/* FORM_URL_DIRNAME - form the dirname of the URL path component */
/*****************************************************************/

static void form_url_dirname(char *http_cmd, char *dir)
{
    int ovector[OV_LEN+1];	/* vector for substring information */
    int match_ret;              /* pcre_exec return code */
    int code;			/* 1st matched substring code */
    int code2;			/* 2nd matched substring code */
    const char *substr;		/* 1st matched substring or NULL */
    const char *substr2;	/* 2nd matched substring or NULL */
    const char *url;		/* the URL part of the log entry */
    char *tmp;			/* temp string */
    int try_again;		/* 1 ==> keep trying the same cycle */

    /* firewall */
    if (http_cmd == NULL || dir == NULL) {
	if (pcredbg) {
	    fprintf(stderr, "form_url_dirname: bail #0\n");
	}
	return;
    }
    dir[0] = '\0'; /* in case we must bail */
    if (pcredbg) {
	fprintf(stderr, "form_url_dirname: http_cmd: %s\n", http_cmd);
    }

    /*
     * extract the URL from the log entry
     */
    memset(ovector, 0, sizeof(ovector));
    match_ret = pcre_exec(pc[RE_GET_URL].re, pc[RE_GET_URL].pe,
			  http_cmd, strlen(http_cmd), 0, 0, ovector, OV_LEN);
    /* look for a match */
    code = 0;
    substr = NULL;
    if (match_ret >= 2) {
	code = pcre_get_substring(http_cmd, ovector, match_ret, 1, &substr);
    }
    /* no match, use the entire http command */
    if (code <= 0 || match_ret < 2 || substr == NULL || substr[0] == '\0') {
	substr = strdup(http_cmd);
	if (substr == NULL) {
	    if (pcredbg) {
		fprintf(stderr, "form_url_dirname: bail #1\n");
	    }
	    return; /* bail */
	}
    }
    url = substr;
    /* url string has been allocated now */
    if (pcredbg) {
	fprintf(stderr, "form_url_dirname: url: %s\n", url);
    }

    /*
     * force a leading / if it does not have one
     */
    if (url[0] != '/') {
	tmp = malloc(1+strlen(url)+1+1);
	if (tmp == NULL) {
	    if (pcredbg) {
		fprintf(stderr, "form_url_dirname: bail #2\n");
	    }
	    return; /* bail */
	}
	strcpy(tmp, "/");
	strcat(tmp, url);
	pcre_free((void *)url);
	url = tmp;
	if (pcredbg) {
	    fprintf(stderr, "form_url_dirname: forced leading /: %s\n", url);
	}
    }

    /* 
     * strip off ? args and fields after whitespace 
     */
    memset(ovector, 0, sizeof(ovector));
    match_ret = pcre_exec(pc[RE_STRIP_ARGS].re, pc[RE_STRIP_ARGS].pe,
			  url, strlen(url), 0, 0, ovector, OV_LEN);
    /* look for a match */
    code = 0;
    substr = NULL;
    if (match_ret == 2) {
	code = pcre_get_substring(url, ovector, match_ret, 1, &substr);
    }
    /* if match, just use URL path (before any ? or whitespace) */
    if (code >= 0 && substr != NULL && substr[0] != '\0') {
	pcre_free((void *)url);
	url = substr;
	if (pcredbg) {
	    fprintf(stderr, "form_url_dirname: stripped url: %s\n", url);
	}
    }

    /*
     * convert //'s into /
     */
    do {
	/* // pattern match */
	memset(ovector, 0, sizeof(ovector));
	match_ret = pcre_exec(pc[RE_UN_SLASHLASH].re, pc[RE_UN_SLASHLASH].pe,
			      url, strlen(url), 0, 0, ovector, OV_LEN);
	/* look for a match */
	try_again = 0;
	code = 0;
	code2 = 0;
	substr = NULL;
	substr2 = NULL;
	if (match_ret == 3) {
	    /* before the // */
	    code = pcre_get_substring(url, ovector, match_ret, 1, &substr);
	    /* after the // */
	    code2 = pcre_get_substring(url, ovector, match_ret, 2, &substr2);
	}
	/* match, remove the foo/../ */
	if (code >= 0 && code2 >= 0 && substr != NULL && substr2 != NULL) {
	    tmp = malloc(strlen(substr)+1+strlen(substr2)+1+1);
	    if (tmp == NULL) {
		if (pcredbg) {
		    fprintf(stderr, "form_url_dirname: bail #3\n");
		}
		return; /* bail */
	    }
	    strcpy(tmp, substr);
	    strcat(tmp, "/");
	    strcat(tmp, substr2);
	    pcre_free((void *)substr);
	    substr = NULL;
	    pcre_free((void *)substr2);
	    substr2 = NULL;
	    pcre_free((void *)url);
	    url = tmp;
	    try_again = 1;
	    if (pcredbg) {
		fprintf(stderr, "form_url_dirname: //-collapse: %s\n", url);
	    }
	}
    } while (try_again != 0);

    /*
     * replace /./ with /
     */
    do {
	/* /./ pattern match */
	memset(ovector, 0, sizeof(ovector));
	match_ret = pcre_exec(pc[RE_SLASHDOT].re, pc[RE_SLASHDOT].pe,
			      url, strlen(url), 0, 0, ovector, OV_LEN);
	/* look for a match */
	try_again = 0;
	code = 0;
	code2 = 0;
	substr = NULL;
	substr2 = NULL;
	if (match_ret == 4) {
	    /* before the /./ */
	    code = pcre_get_substring(url, ovector, match_ret, 1, &substr);
	    /* after the /./ */
	    code2 = pcre_get_substring(url, ovector, match_ret, 3, &substr2);
	}
	/* match, remove the /./ */
	if (code >= 0 && code2 >= 0 && substr != NULL && substr2 != NULL) {
	    tmp = malloc(strlen(substr)+1+strlen(substr2)+1+1);
	    if (tmp == NULL) {
		if (pcredbg) {
		    fprintf(stderr, "form_url_dirname: bail #4\n");
		}
		return; /* bail */
	    }
	    strcpy(tmp, substr);
	    strcat(tmp, substr2);
	    pcre_free((void *)substr);
	    substr = NULL;
	    pcre_free((void *)substr2);
	    substr2 = NULL;
	    pcre_free((void *)url);
	    url = tmp;
	    try_again = 1;
	    if (pcredbg) {
		fprintf(stderr, "form_url_dirname: slashdot: %s\n", url);
	    }
	}
    } while (try_again != 0);

    /*
     * remove foo/../, where foo does not end in .
     */
    do {
	/* foo/../ pattern match */
	memset(ovector, 0, sizeof(ovector));
	match_ret = pcre_exec(pc[RE_UN_DOTDOT].re, pc[RE_UN_DOTDOT].pe,
			      url, strlen(url), 0, 0, ovector, OV_LEN);
	/* look for a match */
	try_again = 0;
	code = 0;
	code2 = 0;
	substr = NULL;
	substr2 = NULL;
	if (match_ret == 3) {
	    /* before the foo/../ */
	    code = pcre_get_substring(url, ovector, match_ret, 1, &substr);
	    /* after the foo/../ */
	    code2 = pcre_get_substring(url, ovector, match_ret, 2, &substr2);
	}
	/* match, remove the foo/../ */
	if (code >= 0 && code2 >= 0 && substr != NULL && substr2 != NULL) {
	    tmp = malloc(strlen(substr)+1+strlen(substr2)+1+1);
	    if (tmp == NULL) {
		if (pcredbg) { 
		    fprintf(stderr, "form_url_dirname: bail #4\n");
		}
		return; /* bail */
	    }
	    strcpy(tmp, substr);
	    strcat(tmp, substr2);
	    pcre_free((void *)substr);
	    pcre_free((void *)substr2);
	    pcre_free((void *)url);
	    url = tmp;
	    try_again = 1;
	    if (pcredbg) {
		fprintf(stderr, "form_url_dirname: ..-reduced: %s\n", url);
	    }
	}
    } while (try_again != 0);

    /*
     * remove trailing /.. from the path
     */
    do {
	/* /foo/.. pattern match */
	memset(ovector, 0, sizeof(ovector));
	match_ret = pcre_exec(pc[RE_TRAIL_DOTDOT].re, pc[RE_TRAIL_DOTDOT].pe,
			      url, strlen(url), 0, 0, ovector, OV_LEN);
	/* look for a match */
	try_again = 0;
	code = 0;
	substr = NULL;
	if (match_ret == 2) {
	    /* before the /.. */
	    code = pcre_get_substring(url, ovector, match_ret, 1, &substr);
	}
	/* match, remove the trailing /.. */
	if (code >= 0 && substr != NULL) {
	    pcre_free((void *)url);
	    url = substr;
	    try_again = 1;
	    if (pcredbg) {
		fprintf(stderr, "form_url_dirname: trailing/..: %s\n", url);
	    }
	}
    } while (try_again != 0);

    /*
     * remove leading /..'s
     */
    do {
	/* leading /../ pattern match */
	memset(ovector, 0, sizeof(ovector));
	match_ret = pcre_exec(pc[RE_LEAD_DOTDOT].re, pc[RE_LEAD_DOTDOT].pe,
			      url, strlen(url), 0, 0, ovector, OV_LEN);
	/* look for a match */
	try_again = 0;
	code2 = 0;
	substr = NULL;
	substr2 = NULL;
	if (match_ret == 3) {
	    /* after the /../'s */
	    code2 = pcre_get_substring(url, ovector, match_ret, 2, &substr2);
	}
	/* match, remove the foo/../ */
	if (code2 >= 0 && substr2 != NULL) {
	    pcre_free((void *)url);
	    url = substr2;
	    try_again = 1;
	    if (pcredbg) {
		fprintf(stderr,
		    "form_url_dirname: leading/.. or /.: %s\n", url);
	    }
	}
    } while (try_again != 0);

    /*
     * remove foo./../
     */
    do {
	/* foo./../ pattern match */
	memset(ovector, 0, sizeof(ovector));
	match_ret = pcre_exec(pc[RE_UN_DOTDOT2].re, pc[RE_UN_DOTDOT2].pe,
			      url, strlen(url), 0, 0, ovector, OV_LEN);
	/* look for a match */
	try_again = 0;
	code = 0;
	code2 = 0;
	substr = NULL;
	substr2 = NULL;
	if (match_ret == 3) {
	    /* before the foo./../ */
	    code = pcre_get_substring(url, ovector, match_ret, 1, &substr);
	    /* after the foo./../ */
	    code2 = pcre_get_substring(url, ovector, match_ret, 2, &substr2);
	}
	/* match, remove the foo./../ */
	if (code >= 0 && code2 >= 0 && substr != NULL && substr2 != NULL) {
	    tmp = malloc(strlen(substr)+1+strlen(substr2)+1+1);
	    if (tmp == NULL) {
		if (pcredbg) {
		    fprintf(stderr, "form_url_dirname: bail #5\n");
		}
		return; /* bail */
	    }
	    strcpy(tmp, substr);
	    strcat(tmp, substr2);
	    pcre_free((void *)substr);
	    substr = NULL;
	    pcre_free((void *)substr2);
	    substr2 = NULL;
	    pcre_free((void *)url);
	    url = tmp;
	    try_again = 1;
	    if (pcredbg) {
		fprintf(stderr, "form_url_dirname: ..-reduced2: %s\n", url);
	    }
	}
    } while (try_again != 0);

    /*
     * a lone /. becomes just /
     */
    if (strcmp(url, "/.") == 0) {
    	tmp = strdup("/");
	if (tmp == NULL) {
	    if (pcredbg) {
		fprintf(stderr, "form_url_dirname: bail #6\n");
	    }
	    return; /* bail */
	}
	pcre_free((void *)url);
	url = tmp;
	if (pcredbg) {
	    fprintf(stderr, "form_url_dirname: lone /.: %s\n", url);
	}
    }

    /*
     * load the modified url into the dir field
     */
    strncpy(dir, url, MAXURL);
    dir[MAXURL] = '\0';
    pcre_free((void *)url);
    url = NULL;
    
    /*
     * strip off the final /element
     *
     * NOTE: We consider the dirname of /chongo/ to be /chongo because
     *	     the URL path /chongo/ really refers to /chongo/index.html
     *	     and the dirname of /chongo/index.html is /chongo
     */
    tmp = strrchr(dir, '/');
    if (tmp != NULL) {
	*tmp = '\0';
    }
    /* unless it was only / */
    if (dir[0] == '\0') {
    	strcpy(dir, "/");
    }
    if (pcredbg) {
	fprintf(stderr, "form_url_dirname: dir: %s\n", dir);
    }
    return;
}
